import numpy as np
import math
import random
from time import time
from scipy.fftpack import dct, idct
import holoviews as hv
import DCT_real_mod
import DCT_shortsupp_mod
hv.extension('bokeh')
renderer = hv.renderer('bokeh')

# Input for recreating the figures in "Real Sparse Fast DCT for Vectors with Short Support" by Sina Bittens and Gerlind
#  Plonka.
N = 2**20
tests = 1000
epsilon = 1e-4
variant = 'exact'
# Input exact data
m = np.array([10, 100, 1000, 10000, 50000])
M = np.array([30, 300, 3000, 30000, 150000])
# Input noisy data
# m = np.array([100, 1000])
# M = np.array([300, 3000])


def test_dct_shortsupp(tests, m, M, N, epsilon, variant='exact'):
    """
    Computes the average runtime and reconstruction error of DCT_real, DCT_shortsupp and idct.
    :param tests: Number of tries
    :param m: support length
    :param M: upper bound on the support length
    :param N: vector length
    :param epsilon: noise threshold
    :param variant: variant=='exact': exact data; variant=='noisy': noisy data (uniform noise)
    :return:
    """

    m = m.ravel()
    M = M.ravel()
    MM = M.shape[0]
    # N needs to be a power of 2, and m, M need to be arrays of the same shape with m <= M componentwise.
    if abs(math.log(N, 2) - round(math.log(N, 2))) < np.spacing(1) and m.shape == M.shape and sum(m <= M) == MM:
        # Initialize the matrix for the average runtimes.
        time_av = np.zeros((MM, 7))
        time_av[:, 0:2] = np.hstack((m[:, None], M[:, None]))
        # Define the SNR values for noisy signals.
        SNR = np.arange(0, 60, 10)
        l_SNR = SNR.shape[0]
        # Define the cutoff thresholds for noisy data, optimized for N=2^20. Initially, use the optimized thresholds
        # for 100 for every support length except 1000. These need to be adapted for support lengths other than 100
        # and 1000.
        val_100 = np.array([2.5, 2, 1, 0.4, 0.15, 0.05])
        val_1000 = np.array([2.5, 2.1, 1.5, 0.85, 0.2, 0.1])
        eps_noise_shortsupp = np.array([2.5, 1.8, 1, 0.3, 0.15, 0.05])

        # Initialize the matrix of reconstruction errors for exact data.
        error_ex = np.zeros((MM, 4))
        plot_list = []
        for j in range(MM):
            # Initialize the matrix for the average runtimes for noisy data.
            time_noise = np.zeros((l_SNR, 5))
            time_noise[:, 0] = SNR
            # Initialize the matrix for the average noisy reconstruction errors.
            error = np.zeros((l_SNR, 5))
            error[:, 0] = SNR
            # Initialize the matrix for the maximal noisy reconstruction errors.
            max_error = np.zeros((l_SNR, 5))
            max_error[:, 0] = SNR
            # Initialize the matrix for the rate of correct recovery.
            correct_recov = np.zeros((l_SNR, 7))
            correct_recov[:, 0] = SNR

            if m[j] < 1000:
                epsilon_noise = val_100
            else:
                epsilon_noise = val_1000

            for t in range(tests):
                if t % 10 == 1:
                    print('{}st test'.format(t))
                # For each test create a random first support index mu for x, which has to be a vector of length N with
                # short support of length m[j].
                mu = random.randint(0, N - m[j])
                x = np.zeros(N)
                # Generate a random number between 0 and floor((m[j] - 2)/2) indices inside the support of x, excluding
                # the first and last support index, whose corresponding entries will be set to zero.
                supp_zero = random.sample(range(mu + 1, mu + m[j] - 1), random.randint(0, math.floor((m[j] - 2) / 2)))
                # Generate uniformly distributed entries between 0 and 10. The entries corresponding to the first and
                # last support index have to be significantly large, i.e., > epsilon and > epsilon_noise.
                if variant == 'exact':
                    x[np.array([mu, mu + m[j] - 1])] = epsilon + (10 - epsilon) * np.random.random(2)
                elif variant == 'noisy':
                    x[np.array([mu, mu + m[j] - 1])] = epsilon_noise[j] + (10 - epsilon_noise[-1]) * np.random.random(2)
                x[np.array([k for k in range(mu + 1, mu + m[j] - 1) if k not in supp_zero])] \
                    = np.random.random(m[j] - len(supp_zero) - 2) * 10

                # Generate the input data.
                dct_x = dct(x, type=2, norm='ortho')

                # Standard IDCT-II Algorithm
                # Measure the runtime of idct.
                start = time()
                S_idct = idct(dct_x, type=2, norm='ortho')
                time_av[j, 5] += time() - start

                # Numerical Experiments for Exact Input Data
                if variant == 'exact':
                    # DCT_real
                    # Measure the runtime of DCT_real for exactly known support lengths.
                    start = time()
                    S_real_ex = DCT_real_mod.DCT_real(dct_x, epsilon, m[j])[0]
                    time_av[j, 2] += time() - start

                    # Measure the runtime of DCT_real for bounded support lengths.
                    start = time()
                    S_real_bound = DCT_real_mod.DCT_real(dct_x, epsilon, M[j])[0]
                    time_av[j, 3] += time() - start

                    # DCT_shortsupp
                    # Measure the runtime of DCT_shortsupp.
                    start = time()
                    S_shortsupp = DCT_shortsupp_mod.DCT_shortsupp(dct_x, epsilon)[0]
                    time_av[j, 4] += time() - start

                    # Compute the reconstruction errors for exact data for the four methods, given by the 2-norm of the
                    # difference between the original vector x and the returned vectors S_real_ex, S_real_bound,
                    # S_shortsupp and S_idct.
                    error_ex[j, 0] += np.linalg.norm(S_real_ex - x) / N
                    error_ex[j, 1] += np.linalg.norm(S_real_bound - x) / N
                    error_ex[j, 2] += np.linalg.norm(S_shortsupp - x) / N
                    error_ex[j, 3] += np.linalg.norm(S_idct - x) / N

                # Numerical Experiments for Noisy Input Data
                elif variant == 'noisy':
                    for n in range(l_SNR):
                        # Generate uniformly distributed noise for the input vector dct_x.
                        noise = np.random.random(N) - 0.5
                        # Scale the noise according to the current SNR value SNR[n].
                        noise = np.linalg.norm(dct_x) * 10**(-SNR[n] / 20) / np.linalg.norm(noise) * noise
                        # Create noisy input data by addition.
                        dct_x_noise = dct_x + noise

                        # Standard IDCT-II Algorithm
                        # Measure the runtime of idct for noisy input data.
                        start = time()
                        S_idct_noise = idct(dct_x_noise, type=2, norm='ortho')
                        time_noise[n, 4] += time() - start

                        # DCT_real
                        # Measure the runtime of DCT_real for exactly known support length and noisy input data.
                        start = time()
                        S_real_noise_ex, mu_noise, m_noise = DCT_real_mod.DCT_real(dct_x_noise, epsilon_noise[n], m[j])
                        time_noise[n, 1] += time() - start
                        # Check whether the support of x was found correctly. First, examine whether the support of x is
                        # contained in the returned support of S_real_noise_ex.
                        if mu >= mu_noise and mu + m[j] - 1 <= mu_noise + m_noise - 1:
                            correct_recov[n, 1] += 1
                            # We also check whether the returned support length of S_real_noise_ex is at most three
                            # times the support length of x.
                            if m_noise <= 3 * m[j] and m_noise >= m[j]:
                                correct_recov[n, 2] += 1

                        # Measure the runtime of DCT_real for bounded short support and noisy input data.
                        start = time()
                        S_real_noise_bound, mu_noise, m_noise \
                            = DCT_real_mod.DCT_real(dct_x_noise, epsilon_noise[n], M[j])
                        time_noise[n, 2] += time() - start
                        # Check whether the support of x was found correctly. First, examine whether the support of x is
                        # contained in the returned support of S_real_noise_bound.
                        if mu >= mu_noise and mu + m[j] - 1 <= mu_noise + m_noise - 1:
                            correct_recov[n, 3] += 1
                            # We also check whether the returned support length of S_real_noise_bound is at most three
                            # times the support length of x.
                            if m_noise <= 3 * m[j] and m_noise >= m[j]:
                                correct_recov[n, 4] += 1

                        # DCT_shortsupp
                        # Measure the runtime of DCT_shortsupp for noisy input data.
                        start = time()
                        S_shortsupp_noise, struc, mu_noise, m_noise = \
                            DCT_shortsupp_mod.DCT_shortsupp(dct_x_noise, eps_noise_shortsupp[n])
                        time_noise[n, 3] += time() - start
                        # Check whether the support of x was found correctly. First examine whether the support of x is
                        # contained in the returned support of S_shortsupp_noise.
                        if mu_noise + m_noise <= N-1:
                            if mu >= mu_noise and mu + m[j] - 1 <= mu_noise + m_noise - 1:
                                correct_recov[n, 5] += 1
                                # We also check whether the returned support length of S_shortsupp_noise is at most
                                # three times the length of x.
                                if m_noise <= 3 * m[j] and m_noise >= m[j]:
                                    correct_recov[n, 6] += 1
                        else:
                            # If 0, N-1 are contained in the support of S_shortsupp_noise, the block has to be
                            # considered modulo N.
                            if (mu >= mu_noise or mu <= mu_noise + m_noise - 1 - N) and \
                                    (mu + m[j] - 1 >= mu_noise or mu + m[j] - 1 <= mu_noise + m_noise - 1 - N):
                                if m_noise <= 3 * m[j] and m_noise >= m[j]:
                                    correct_recov[n, 6] += 1

                        # Compute the reconstruction errors for noisy data for the four methods.
                        norm_1 = np.linalg.norm(S_real_noise_ex - x) / N
                        norm_2 = np.linalg.norm(S_real_noise_bound - x) / N
                        norm_3 = np.linalg.norm(S_shortsupp_noise - x) / N
                        norm_4 = np.linalg.norm(S_idct_noise - x) / N
                        error[n, 1] += norm_1
                        error[n, 2] += norm_2
                        error[n, 3] += norm_3
                        error[n, 4] += norm_4
                        # Compute the maximal reconstruction errors for noisy data for the four methods.
                        if norm_1 > max_error[n, 1]:
                            max_error[n, 1] = norm_1
                        if norm_2 > max_error[n, 2]:
                            max_error[n, 2] = norm_2
                        if norm_3 > max_error[n, 3]:
                            max_error[n, 3] = norm_3
                        if norm_4 > max_error[n, 4]:
                            max_error[n, 4] = norm_4

            # Compute the average noisy runtimes and reconstruction errors.
            if variant == 'noisy':
                time_noise[:, 1:4] = time_noise[:, 1:4] / tests
                error[:, 1:4] = error[:, 1:4] / tests
                correct_recov[:, 1:6] = correct_recov[:, 1:6] / tests

                # Create Plots
                # Average Reconstruction Errors
                lines_real_ex = hv.Curve((error[:, 0], error[:, 1]), label="DCT_real (exact support)").options(
                    color='cyan')
                points_real_ex = hv.Points((error[:, 0], error[:, 1])).options(color='cyan', size=8, marker='^')
                lines_real_bound = hv.Curve((error[:, 0], error[:, 2]), label="DCT_real (bounded support)").options(
                    color='blue')
                points_real_bound = hv.Points((error[:, 0], error[:, 2])).options(color='blue', size=8,
                                                                                  marker='v')
                lines_shortsupp = hv.Curve((error[:, 0], error[:, 3]), label="DCT_shortsupp").options(
                    color='red')
                points_shortsupp = hv.Points((error[:, 0], error[:, 3])).options(color='red', size=8,
                                                                                 marker='s')
                lines_idct = hv.Curve((error[:, 0], error[:, 4]), label="idct_2").options(color='lime')
                points_idct = hv.Points((error[:, 0], error[:, 4])).options(color='lime', size=8, marker='d')
                real_ex = (lines_real_ex * points_real_ex)
                real_bound = (lines_real_bound * points_real_bound)
                shortsupp = (lines_shortsupp * points_shortsupp)
                idct_plot = (lines_idct * points_idct)

                plot_av_err = (real_ex * real_bound * shortsupp * idct_plot)\
                    .options(logy=True, xlabel='SNR', ylabel='Average Reconstruction Error', legend_position='right',
                             width=750, height=300, show_grid=True)\
                    .relabel('Average Reconstruction Error, m = {}'.format(m[j]))
                renderer.save(plot_av_err, 'average_error_DCT_{}'.format(m[j]))
                plot_list.append(plot_av_err)

                # Maximal Reconstruction Errors
                lines_real_ex = hv.Curve((max_error[:, 0], max_error[:, 1]), label="DCT_real (exact support)").options(
                    color='cyan')
                points_real_ex = hv.Points((max_error[:, 0], max_error[:, 1])).options(color='cyan', size=8, marker='^')
                lines_real_bound = hv.Curve((max_error[:, 0], max_error[:, 2]), label="DCT_real (bounded support)")\
                    .options(color='blue')
                points_real_bound = hv.Points((max_error[:, 0], max_error[:, 2])).options(color='blue', size=8,
                                                                                          marker='v')
                lines_shortsupp = hv.Curve((max_error[:, 0], max_error[:, 3]), label="DCT_shortsupp").options(
                    color='red')
                points_shortsupp = hv.Points((max_error[:, 0], max_error[:, 3])).options(color='red', size=8,
                                                                                         marker='s')
                lines_idct = hv.Curve((max_error[:, 0], max_error[:, 4]), label="idct_2").options(color='lime')
                points_idct = hv.Points((max_error[:, 0], max_error[:, 4])).options(color='lime', size=8, marker='d')
                real_ex = (lines_real_ex * points_real_ex)
                real_bound = (lines_real_bound * points_real_bound)
                shortsupp = (lines_shortsupp * points_shortsupp)
                idct_plot = (lines_idct * points_idct)

                plot_max_err = (real_ex * real_bound * shortsupp * idct_plot)\
                    .options(logy=True, xlabel='SNR', ylabel='Maximal Reconstruction Error', legend_position='right',
                             width=750, height=300, show_grid=True)\
                    .relabel('Maximal Reconstruction Error, m = {}'.format(m[j]))
                renderer.save(plot_max_err, 'maximal_error_DCT_{}'.format(m[j]))
                plot_list.append(plot_max_err)

        # Compute the average runtimes and the average reconstruction errors.
        time_av[:, 2:] = time_av[:, 2:] / tests
        print(time_av)
        error_ex = error_ex / tests
        print('error_ex={}'.format(error_ex))

        # Create Plots
        if variant == 'exact':
            # Runtime
            lines_real_ex = hv.Curve((time_av[:, 0], time_av[:, 2]), label="DCT_real (exact support)")\
                .options(color='cyan')
            points_real_ex = hv.Points((time_av[:, 0], time_av[:, 2])).options(color='cyan', size=8, marker='^')
            lines_real_bound = hv.Curve((time_av[:, 0], time_av[:, 3]), label="DCT_real (bounded support)")\
                .options(color='blue')
            points_real_bound = hv.Points((time_av[:, 0], time_av[:, 3])).options(color='blue', size=8, marker='v')
            lines_shortsupp = hv.Curve((time_av[:, 0], time_av[:, 4]), label="DCT_shortsupp").options(
                color='red')
            points_shortsupp = hv.Points((time_av[:, 0], time_av[:, 4])).options(color='red', size=8, marker='s')
            lines_idct = hv.Curve((time_av[:, 0], time_av[:, 5]), label="idct_2").options(color='lime')
            points_idct = hv.Points((time_av[:, 0], time_av[:, 5])).options(color='lime', size=8, marker='d')
            real_ex = (lines_real_ex * points_real_ex)
            real_bound = (lines_real_bound * points_real_bound)
            idct_plot = (lines_idct * points_idct)
            shortsupp = (lines_shortsupp * points_shortsupp)

            plot_runtime = (real_ex * real_bound * shortsupp * idct_plot)\
                .options(logx=True, logy=True, xlabel='Support Length m', ylabel='Runtime [s]', legend_position='right',
                         width=750, height=300, show_grid=True).relabel('Average Runtime')
            renderer.save(plot_runtime, 'runtime_DCT')

        elif variant == 'noisy':
            plot_errors = hv.Layout(plot_list).cols(2)
            renderer.save(plot_errors, 'errors_DCT')

    else:
        print('One of the parameters has been set incorrectly.')


test_dct_shortsupp(tests, m, M, N, epsilon, variant)